/* 
* ISG - Infopipe Stub Generator
*  version 0.2 - started using XSLT in the processing
*
*  Galen Swint(zorn@cc.gatech.edu)
*  Younggyun Koh(young@cc.gatech.edu)
*
*/

#include <stdio.h>
#include <string.h>
#include <unistd.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fstream.h>

#include "isg.hpp"
#include "out.hpp"
#include "domErrorHandler.hpp"
#include "isgUtil.hpp"
#include "isgXslt.hpp"

unsigned int gFlags;

void usage()
{
    cerr << "isg: Infopipe Stub Generator" << endl;
    cerr << "Usage: isg [-l] -s spec.xip -t template.xsl" << endl;
    cerr << " -s filename		xip file" << endl;
    cerr << " -t filename	xslt template file" << endl;
    cerr << " -l directory	xslt template directory" << endl;
    //cerr << " -d [filename]		dump list of nodes to the specified file or stdout" << endl;
    //cerr << " -e                        dump template file to stdout" << endl;
    //cerr << " -f                        publish node set document in the file system" << endl;
    cerr << endl;

    cerr << " If needed, set ISG_TEMPLATE_PATH to the (single) directory" << endl;
    cerr << " containing the templates." << endl << endl;
    cerr << " see http://www.cc.gatech.edu/projects/infosphere";
    cerr << " for more information" << endl << endl;
    exit(1);
}

int parseOptions( int argc, char **argv, char **specName, char **templateName, char **templatePath, char **dumpFileName)
{

  int c;

  gFlags = 0;

  while((c = getopt(argc, argv, "s:t:l:dcef")) != -1)
    switch(c)
      {
      case 's': // infopipe spec file name
        if (optarg == NULL)
          gFlags |= FLAG_ERROR;
        else
        {
          *specName = optarg;
          gFlags |= FLAG_SPEC;
        }
        break;

      case 'l': // template path 
        if (optarg == NULL)
          gFlags |= FLAG_ERROR;
        else
        {
          *templatePath = optarg;
          gFlags |= FLAG_SPEC;
        }
	      break;

      case 't': // template file name
        if (optarg == NULL)
          gFlags |= FLAG_ERROR;
        else
        {
          *templateName = optarg;

          // for multiple templates
          //cout << "templates: " << optarg;
          while(optind < argc && argv[optind][0] != '-')
            cout << " " << argv[optind++];
          cout << endl;
          
          gFlags |= FLAG_TEMPLATE;
        }
        break;

      case 'd': // dump mode
        gFlags |= FLAG_DUMP_NODESET;
        if (optind < argc && argv[optind][0] != '-')
          *dumpFileName = argv[optind++];
        else
          *dumpFileName = NULL;

        break;

      case 'e': 
	gFlags |= FLAG_DUMP_TEMPLATE;
	break;
      case 'f':
	gFlags |= FLAG_PUBLISH_FILE;
	break;
      case '?':
        cout << "Invalid option(s)" << endl;
      default:
        usage();
      }

  if (gFlags & FLAG_ERROR)
    usage();

  // if there is neither of spec and template files
  if (!((gFlags & FLAG_SPEC) && (gFlags & FLAG_TEMPLATE)))
    usage();

  return 0;
}

int initializeXML( void );
DOM_Document getDomDoc( const char *filename );
int parseFile(DOM_Document *domDoc, char *filename, char *dir, ...);

int oldIsgProcessing(DOM_Document templateDocument, DOM_Document specDocument, char *dumpFileName);

int main( int argc, char **argv )
{
  char *xipName;
  char *templateName; 
  char *dumpFileName;
  char *templatePath;

  /******************/
  /* Initialization */
  /******************/

  cout << "Welcome to the Infopipe Stub Generator." << endl;

  initializeXML( );
  initializeXSLTProcessor();

  parseOptions(argc, argv, &xipName, &templateName, &templatePath, &dumpFileName);

  /*************/
  /*  Parsing  */
  /*************/

  DOM_Document xipDocument;
  DOM_Document templateDocument;
  DOM_Document outputDocument;

  cout << "isg: parsing template file...";

  char foundPath[128];
  
  /* parse template file */
  if (parseFile(&templateDocument, templateName, foundPath, templatePath, getenv("ISG_TEMPLATE_PATH")))
  { 
    cout << "isg: error while parsing template file '" 
         << templateName << "'." << endl;
    exit(1);
  }

  cout << "done" << endl;

  cout << "isg: parsing xip file...";
  
  /* parse xip file */
  if (parseFile(&xipDocument, xipName, NULL, getenv("ISG_XIP_PATH")))
  {
    cout << "isg: error while parsing xip file '" 
         << xipName << "'." << endl;
    exit(1);
  }  

  cout << "done" << endl;

  // change current dir to template path, to work well with xsl:include
  chdir(foundPath);

  cout << "isg: doing pre-processing...un-implemented" << endl;

  cout << "isg: doing xslt processing...";
  
  /* XSLT processing */
  if (doXSLTstep(xipDocument, templateDocument, outputDocument))
  {
    cout << "isg: error while xslt processing" << endl;
    exit(1);
  }
  
  cout << "done" << endl;

  /* old ISG processing */
  cout << "isg: doing post-processing..." << flush;
  oldIsgProcessing(outputDocument, xipDocument, dumpFileName);
  cout << "done" << endl;

  /* print out the output directory */
  char cwd[128];
  getcwd(cwd, sizeof(cwd));
  cout << "isg: writing files into " << cwd << endl;

  /* post processing */
  writeFiles(outputDocument);

  return 0;
}

/* 
 parseFile() - load and parse XML document named filename
*/
int parseFile(DOM_Document *domDoc, char *filename, char *dir, ...)
{
  // current directory in which we are looking for the template
  char *curDir = "./";
  // full path to the template we want
  char *fullPath;
  struct stat fileStatus;
  char *thisFunc = "parseFile()";

  va_list ap;

  // for each item in list of args
  va_start(ap, filename);
  do
  {
    // build the full path
    fullPath = new char[strlen(curDir) + strlen(filename) + 2];
    // copy in directory base
    strcpy(fullPath, curDir);

    // add a trailing slash if needed
    if( !(fullPath[ strlen(fullPath) - 1 ] == '/' ) )
      strcat(fullPath, "/");

    // now add filename
    strcat(fullPath, filename);

    // check to see if there
    if (!stat(fullPath, &fileStatus))
    {
      // file is there
      va_end(ap);

      // error check for directory, etc.
      if (!S_ISREG(fileStatus.st_mode))
      {
        cout << fullPath << " is not a regular file." << endl;
        delete fullPath;
        return 1;
      }
      
      // parse the file
      *domDoc = getDomDoc(fullPath);

      // return the directory in which the file is found
      if (dir) strcpy(dir, curDir);

      // cleanup and done
      delete fullPath;
      return 0;
    }
    delete fullPath;
  }
  // iterate to next argument provided
  while(curDir = va_arg(ap, char *));
  va_end(ap);
  
  cout << "Can't find file " << filename << " does not exist." << endl;
  return 1;
}

int initializeXML( void )
{
  // initialize XML4C
  try
    {
    XMLPlatformUtils::Initialize();
    }
  catch( const XMLException& caught )
    {
    cerr << "***Error initializing XML4C library:"  << endl;
    // StrX() doesn't seem to work. Don't know why.
    //  << StrX( caught.getMessage() ) << endl;
    exit( 1 );
    }
  return 0;
}
  
/*
 getDomDoc() - takes file name of XML document and returns DOM_Document
*/
DOM_Document getDomDoc( const char *filename )
{
  DOMParser dp;
  domErrorHandler isgErrorHandler;
  int ecount;

  // needed for Xalan to accept DOM tree
	//dp.setToCreateXMLDeclTypeNode(false);
  //dp.setDoNamespaces(true); // really needed?

  dp.setErrorHandler( &isgErrorHandler );
  dp.setValidationScheme( DOMParser::Val_Auto );
  try 
  {
    dp.parse( filename );
    ecount = dp.getErrorCount();
    if( ecount )
    {
      cout << "isg: Errors in parsing '" << filename << "'." << endl;
      exit(1);
    }
  }
  catch(...)
  {
    cout << "isg: parser generated an exception in parsing '" << filename << "'." << endl;

  }
  return ( dp.getDocument() );
}


