<?xml version="1.0" ?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.0">

<!-- this is a mapping of the infopipe types -->
<xsl:variable name="maptables" select="document('ECho2mapTables.xml')/mapTables"/>

<xsl:template name="filterCode">
 <xsl:param name="thisPipeName"/>
 <xsl:param name="portName"/>

// pipe <xsl:value-of select="$thisPipeName"/>
// port <xsl:value-of select="$portName"/>

#include &lt;string.h&gt;
// filterInfoStructDef -- defines a struct which holds data about
// -- a filter available for installation
// -- name - name by which a filter is refereneced
// -- inType - input type a filter expects, should matching ouput of pipe on which filter is placed
// -- outType - output type a filter expects, could be ""
// -- outTypePtr - output type a filter expects, could be NULL
// -- paramsType - the type (a char name) of params that a filter may use, could be NULL
struct filterInfoStructDef
{
  char *name;
  char *uri;
  char *inType;
  char *outType;
  IOFieldList outTypePtr;  
  int   takesParams;
  IOFieldList paramsTypeList;
};

// filterTableStructDef -- way to store filter info about
// -- available filters. Fields map what was named above.
typedef struct filterInfoStructDef filterInfo;
filterInfo filterTable[] = {

<xsl:for-each select="//filter">
// &lt;block name="filter_<xsl:value-of select="@name"/>"&gt;
  { "<xsl:value-of select="@name"/>",
      "<xsl:value-of select="@uri"/>",    
          "<xsl:value-of select="@inType"/>",     
              "<xsl:value-of select="@outType"/>",  
                  <xsl:value-of select="@outType"/>FieldList,  
                      <xsl:value-of select="@takesParams"/>, 
                          <xsl:choose><xsl:when test="@paramType"><xsl:value-of select="@paramType"/>FieldList</xsl:when> <xsl:otherwise>NULL</xsl:otherwise></xsl:choose> },
// &lt;/block&gt;</xsl:for-each>
  { NULL, NULL, NULL, NULL, NULL, 0, NULL }  };


// getFilterCode() -- pull the code that is the source for a filter. Currently,
// only works for retrieving code from a file. Returns NULL if no source found.
// Also, do NOT free returned buffer. That buffer is owned by this functions.
static char *getFilterCode( char *filterName )
{
  filterInfo *curr;
  char *location = NULL;

  // conserve our buffer allocation
  // create a "grow only" strategy for get larger enough buffer
  static char *filterSourceBuffer;
  // remember buffer size forever
  static int bufferSize = 0;

  // read new each time
  int totalRead = 0, lastRead = 0;

  // FILE ptr we read from
  FILE *srcFile;

  // first, get URI of buffer location, assume file for now
  while( curr-&gt;name != NULL )
  {
    if( !strcmp( curr-&gt;name, filterName ) )
      location = curr-&gt;uri;
  }

  // if can't find URI, return NULL
  if( location == NULL )
    return NULL;

  // open for read ing the sourcefile
  srcFile = fopen( location, "r" );
  if( srcFile == NULL )
  {
    // if can't get file, dump out
    fprintf(stderr, "Can't open filter file.");
    return NULL;
  }


  // read to end of file
  while( !feof( srcFile ) )
  {
    // check to see how much we've read.
    // if total read this time is size of buffer, we need more buffer space,
    // so allocate another 4K (one page)
    if( bufferSize == totalRead )
    {
      bufferSize += 4096;
      filterSourceBuffer = (char *)realloc( filterSourceBuffer, bufferSize );
      if( !filterSourceBuffer )
      {
        fprintf(stderr, "Can't allocate space for buffer source." );
        return NULL;
      }
    }
    lastRead = fread( filterSourceBuffer + totalRead, bufferSize - totalRead, (size_t) 1, srcFile);
    totalRead += lastRead;
  }

  if( totalRead == bufferSize )
  {
    bufferSize += 4096;
    filterSourceBuffer = (char *)realloc( filterSourceBuffer, bufferSize );
    if( !filterSourceBuffer )
    {
      fprintf(stderr, "Can't allocate space for buffer source -- 2." );
      return NULL;
    }
  }

  // null terminate the source
  filterSourceBuffer[ totalRead ] = 0;
  printf( "Source:\n%s\n", filterSourceBuffer );

  // close the file
  fclose(srcFile);

  // done
  return filterSourceBuffer;
}

// getFilterOutType() -- returns the string name of the output type of the 
// -- filter named filterName, NULL if not found in
// -- the list
static char *getFilterOutType( char *filterName )
{
  filterInfo *curr;
  while( curr-&gt;name != NULL )
    {
    if( !strcmp( curr-&gt;name, filterName ) )
      return curr-&gt;outType;
    }
  return NULL;
}

// getFilterOutTypePtr() -- returns a pointer to the output type field list of the 
// -- filter named filterName, NULL if not found in
// -- the list
static IOFieldList getFilterOutTypePtr( char *filterName )
{
  filterInfo *curr;
  while( curr-&gt;name != NULL )
    {
    if( !strcmp( curr-&gt;name, filterName ) )
      return curr-&gt;outTypePtr;
    }
  return NULL;
}


// getFilterInType() -- returns the string name of the input type of the 
// -- filter named filterName, NULL if not found in
// -- the list
static char *getFilterInType( char *filterName )
{
  filterInfo *curr;
  while( curr-&gt;name != NULL )
    {
    if( !strcmp( curr-&gt;name, filterName ) )
      return curr-&gt;inType;
    }
  return NULL;
}

// getFilterParamsTypeList() -- returns the name IOFieldList for the params of the 
// -- filter named filterName, NULL if not found in
// -- the list
static IOFieldList getFilterParamsTypeList( char *filterName )
{
  filterInfo *curr;
  while( curr-&gt;name != NULL )
    {
    if( !strcmp( curr-&gt;name, filterName ) )
      return curr-&gt;paramsTypeList;
    }
  return NULL;
}

// getFilterTakesParams() -- this function looks up whether a filter 
// -- requires parameters be supplied to function correctly. If so,
// -- then the installFilter_<xsl:value-of select="$thisPipeName"/>() function expect that 
// -- params be provided
static int getFilterTakesParams( char *filterName )
{
  filterInfo *curr;
  while( curr-&gt;name != NULL )
    {
    if( !strcmp( curr-&gt;name, filterName ) )
      return curr-&gt;takesParams;
    }
  return -1;
}


// filterCheck( ) -- checks to see if a filter will work
// -- on the current incoming data stream. Primarily, this
// -- is a type check. I.e. a filter is a typed object
// -- and its input type must match the type coming from
// -- our source.
static int filterCheck( char *filterName ) { return 1; }

// we need to keep track of all the filters installed on this channel
// -- In doing so, we need to know the name of a filter, the ID under
// -- which it publishes its data, and any filters on top of which
// -- it may be stacked.
struct filterStackElementStruct
{
  char *filterName;
  char *connToChanId;
  struct filterStackElementStruct *next;
};
typedef struct filterStackElementStruct filterStackElement;
typedef struct filterStackElementStruct *filterStackElementPtr;

// We need to keep track of all filters installed so we
// -- keep a list of them on a stack. This lets us un-install
// -- them later one at a time and we will still know what
// -- type to expect.
static filterStackElementPtr filterStackTop = NULL; 
    
// getFilterStackString() -- returns a string which lists the filters
// -- in the current filter stack. Each is listed by filterName
// -- and filterName's are separated by ':'
char *getFilterStackString(){
  char *buf; // holds result 
  int bufsize;
  int totalLen = 0;
  int partialLen;
  filterStackElementPtr current = filterStackTop;

  // allocate a space for string and term to len 0
  buf = (char *)malloc( sizeof(char) * 100 );
  if( !buf ) {
    fprintf(stderr, "getFilterStackString(): out-of-mem err[2].");
  }
  buf[0] = 0;
  bufsize = 0;
  totalLen = 0;

  while( current ){
    // be sure we have space in buffer
    partialLen = strlen( current-&gt;filterName );
    while( bufsize &lt;= (totalLen + partialLen + 2) ) { 
                                       // plus 2 for NULL term and ':'
      buf = (char *)realloc( buf, 100 );
      if( !buf ) {
        fprintf(stderr, "getFilterStackString(): out-of-mem err[2].");
      }
    }

    // append separator (if not first)
    if( totalLen > 0 )
      strcat( buf, ":" );

    // append string
    strcat( buf, current-&gt;filterName );
  }
  
  return buf;
}




// installFilter_<xsl:value-of select="$thisPipeName"/>() -- adds a filter to the current
// -- incoming data stream. If a filter is already installed, this
// -- will install a filter "stacked" on top of the older filter.
// -- you can query the stack with other functions.
// filterName - name of filter to install
// filterParams - params that may be needed by the filter for it
//              - to function. If no params to be passed then
//              - send null.
// -- returns 0 if successful, non-zero otherwise
int installFilter_<xsl:value-of select="$thisPipeName"/>( char *filterName, void *filterParams )
{
  char lookupStr[512];
  char *filterCode;
  filterStackElementPtr fse;
  EChannel filteredChannel;
  ECdata_struct paramSpec;
  

  // check to be sure the filter will work on this stream
  // (not implemented)
  if(!filterCheck( filterName )) { 
    fprintf(stderr, "Filter incompatible with existing stream type." ); 
    return -1; 
  }


  // add filter to stack of filters
  // allocate for stack
  fse = (filterStackElementPtr) malloc( sizeof( filterStackElement) );
  if( fse == NULL ) { 
    fprintf(stderr, "Error no mem."); 
    return -1; 
  }

  // save the name
  fse-&gt;filterName = (char *)malloc( strlen( filterName ) + 1 );
  if( fse-&gt;filterName == NULL ) {
    fprintf(stderr, "Error no mem."); free( fse ); 
    return -1; 
  }
  strcpy( fse-&gt;filterName, filterName );

  // save the old channelId
  fse-&gt;connToChanId = (char *)malloc( strlen( incomingGlobalID ) + 1 );
  strcpy( fse-&gt;connToChanId, incomingGlobalID );
  
  // stack it up
  fse-&gt;next = filterStackTop;
  filterStackTop = fse;

  // load filter code from URI
  filterCode = getFilterCode(filterName);


  // 3 cases:
  // 1) filter is boolean - output type is input type
  // 2) filter has new output type
  // 3) filter has new output type AND is paramterized


  // *****CASE 1***** boolean filter
  if( !getFilterOutTypePtr( filterName ) ) {
    filteredChannel = EChannel_typed_derive( ec, fse-&gt;connToChanId, filterCode, getFilterOutTypePtr( filterName ), NULL );

  }
  else if( !getFilterTakesParams(filterName) ){
    filteredChannel = EChannel_typed_derive_data( ec, fse-&gt;connToChanId, filterCode, getFilterOutTypePtr( filterName ), NULL, filterParams );
     
  }
  // is it a filter that changes data or is simply boolean?
  else {

    // check to be sure params there
    if( filterParams == NULL )
    {
      fprintf(stderr, "No params supplied. Params needed.\n");
      return 1;
    }
/*
    // allocate param struct
    if( !paramStruct ) {
      paramStruct = (ECdata_spec) malloc( sizeof( ECdata_struct ) );
      if( !paramStruct ) {
        fprintf( stderr, "Unable to allocate mem needed for parameterized filtering.\n");
        return 1;       
      }
    }
*/
    // set up our params struct
    paramSpec.data_field_list = getFilterParamsTypeList( filterName );
    paramSpec.data_subformat_list = NULL;
    paramSpec.initial_value = filterParams;
    // derive
    filteredChannel = EChannel_typed_derive_data( ec, fse-&gt;connToChanId, filterCode, getFilterOutTypePtr( filterName ), NULL, &amp;paramSpec );

  }
  
  ECsink_typed_subscribe( filteredChannel, 
                           getFilterOutTypePtr(filterName),
                            NULL, <xsl:value-of select="$thisPipeName"/>__<xsl:value-of select="$portName"/>Handler, NULL);
<!--
  incomingGlobalID = lookup("<xsl:value-of select="$thisPipeName"/><pipeName/>","<connType/>","<xsl:value-of select="$portName"/><connName/>", PUBLISH_FILE, lookupStr);
  <xsl:value-of select="$thisPipeName"/><pipeName/>_<connType/>_<xsl:value-of select="$portName"/><connName/>ReceiveChannel = EChannel_open( ec,  incomingGlobalID);
-->

  // 0 for success
  return 0;
}

int uninstallLastFilter_<xsl:value-of select="$thisPipeName"/>( void )
{

  // uninstalls the last filter installed
  filterStackElementPtr fse;

  // do nothing if no filter installed at all
  if( filterStackTop == NULL ) return 0;

  // pop from stack
  fse = filterStackTop;
  filterStackTop = filterStackTop-&gt;next;

  // release space
  free( fse-&gt;filterName );
  free( fse-&gt;connToChanId );
  free( fse );

  return 0;
}
<xsl:text>
</xsl:text>
</xsl:template>
</xsl:stylesheet>
